## Phone2SAS version 1.0 ##
## requirements: python3, numpy, open3d
# To run, type "python phone2sas.py y.ply --dmax 160.0". see help (-h).
import open3d as o3d
import numpy as np
import argparse

### User-defined parameters ###
parser = argparse.ArgumentParser()
parser.add_argument("ply_file", help="Path to the 3D-scanned file (.ply)")
parser.add_argument("--dmax", type=float, default=160.0, help="in angstrom; default: 160.0")
parser.add_argument("--rp", type=int, default=10000, help="the number of random points for sampling; default: 10000")
args = parser.parse_args()
print("Path to the 3D-scanned file (.ply): ",args.ply_file)
print("Dmax: ", args.dmax)
print("Number of sampling: ", args.rp)

filename_3d = args.ply_file # file name (path) of 3D-scanned data (e.g., "y.ply")
dmax = args.dmax # Dmax of the object in angstrom
r_num = args.rp # integer. the number of random points (default: 10000)

## box size ##
pcd = o3d.io.read_point_cloud(filename_3d)
pcd_array = np.asarray(pcd.points)
wmax = np.max(pcd_array, axis = 0)
wmin = np.min(pcd_array, axis = 0)
box = wmax - wmin
bx = box[0]; by = box[1]; bz = box[2]
## random points ##
random_pts = np.zeros((r_num,3))
random_pts[:,0] = np.random.rand(r_num)*bx + wmin[0]
random_pts[:,1] = np.random.rand(r_num)*by + wmin[1]
random_pts[:,2] = np.random.rand(r_num)*bz + wmin[2]
## read 3D data by open3D functions ##
mesh = o3d.io.read_triangle_mesh(filename_3d)
mesh = o3d.t.geometry.TriangleMesh.from_legacy(mesh)
scene = o3d.t.geometry.RaycastingScene()
_ = scene.add_triangles(mesh)
## find points inside by open3D functions ##
query_point = o3d.core.Tensor(random_pts, dtype=o3d.core.Dtype.Float32)
occupancy = scene.compute_occupancy(query_point)
## find coordinates of the points inside ##
in_or_not = occupancy.numpy() > 0 # alternatively, "== 0" gives outside points
w2d = random_pts[in_or_not] # extract xyz values with "True"

## pc2pr: point cloud to P(r)
row, col = w2d.shape
num = row
comb_num_2 = int(num*(num - 1) / 2)
temp = np.zeros((comb_num_2,1))

s = 0
for i in range(num - 1):
    tmp_a = w2d[i]
    for h in range(i+1, num):
        tmp_b = w2d[h]
        tmp = tmp_a - tmp_b
        temp[s] = np.linalg.norm(tmp)
        s += 1

num_bin = 101
w_pr, bin_edge = np.histogram(temp, bins=num_bin)
V_scale = bin_edge[num_bin]
wr_edge = dmax*bin_edge/V_scale
wr = ((wr_edge + np.roll(wr_edge, -1))/2)[:-1]

## pr2iq: P(r) to I(q) ##
num_q = 501
iq = np.zeros((num_q-1,1))
q2iq = np.zeros((num_q-1,1))
wq = 0.001*np.array(range(1,num_q,1)) # q = 0.001 - 0.5 A-1
num_wr = num_bin - 1
nakaint = np.zeros((num_wr,1))
pi = np.pi

i = 0
for q in wq:
    nakaint = 4*pi*w_pr*np.sin(q*wr)/(q*wr)
    #nakaint[0] = 0 # use when wt has 0 value.
    sum = 0
    for k in range(1,num_wr):
        dr = wr[k] - wr[k-1]
        sum += nakaint[k]*dr
    iq[i] = sum
    q2iq[i] = q*q*sum
    i+=1

iqvsq = np.c_[wq,iq]
kratky = np.c_[wq,q2iq]
ddf = np.c_[wr,w_pr]

## output ##
#np.savetxt('random_points.txt', random_pts)
np.savetxt('inside_points.txt', w2d)
np.savetxt('iq.txt', iqvsq)
np.savetxt('kratky.txt', kratky)
np.savetxt('pr.txt', ddf)
